md5 : 32556a6d38166e68102b76f14b88b749
name : Bio::SeqAlignment::Components::Libraries::edlib::OpenMP_3255
version : ""
language : C
language_id : C
installed : 0
date_compiled : Thu Jun 13 18:01:34 2024
inline_version : 0.86
ILSM : %
    module : Inline::C
    suffix : so
    type : compiled
Config : %
    apiversion : ?
    archname : x86_64-linux
    cc : cc
    ccflags : -fwrapv -fno-strict-aliasing -pipe -fstack-protector-strong -I/usr/local/include -D_LARGEFILE_SOURCE -D_FILE_OFFSET_BITS=64
    ld : cc
    osname : linux
    osvers : 5.19.0-45-generic
    so : so
    version : 5.38.0
